#------------------------------------------------------------------------
# sddgen: generator.py
#------------------------------------------------------------------------
import re
import logging

logger = logging.getLogger(__name__)

class Generator:
    def __init__(self, name='', model=None, output='sddout'):
        self.name = name
        self.data = model
        #
        self.header = None
        self.document = None
        self.components = None
        self.styles = None
        if self.data:
            self.header = self.data.header
            self.document = self.data.document
            self.components = self.data.components
            self.styles = self.data.styles
        self.outputName = output
        #
        
    def findStyleFor(self, ctype):
        x = None
        if self.styles:
            logging.debug('Styles are there')
            if ctype in self.styles.keys():
                logging.debug('  Style for %s found' % ctype)
                x = self.configuration[ctype]
            else:
                logging.debug('  Style for %s cannot be found' % ctype)
        else:
            logging.debug('  Style is empty')
        return x

    def findComponent(self, typeName):
        component = None
        return component
    
    def writeDocument(self, fout):
        pass
    
    def writeEntry(self, fout):
        pass
    
    def writeStyles(self, fout):
        pass
    
    def writeStyle(self, fout):
        pass
    
    def generate(self):
        cname = self.__class__.__name__
        logger.warning('Cannot generate output for %s' % cname)
        pass

    pass

class TkGenerator(Generator):
    def __init__(self, data):
        super().__init__(name='TkGenerator', data=data)

    def generate(self, fn_out):
        prefix=''
        dashes = '-'*70
        with open(fn_out, 'w') as fout:
            fout.write('#%s\n' % dashes)
            fout.write('# Generated by ssdgen.py from %s\n' % self.filenameIn)
            fout.write('#%s\n' % dashes)
            fout.write('import tkinter as tk\n')
            fout.write('from tkinter import ttk\n')
            fout.write('\n')
            self.outputComponent(self.documentKey, self.document, fout, prefix)
            for compKey, comp in self.components.items():
                compName, tags = self.decodeKey(compKey)
                baseName = ''
                if len(tags)>0: baseName = tags[0]
                fout.write('\n')
                fout.write('# Component: %s\n' % compName)
                self.outputComponent(compKey, comp, fout, prefix)
        pass

    def outputPlacement(self, name, data, fout, prefix):
        logger.debug('Write out placement of %s' % name)
        if data != None:
            if 'pack' in data.keys():
                pack = data['pack']
                s = prefix + '%s.pack(' % name
                opts = ('side', 'anchor', 'fill', 'expand')
                for opt in opts:
                    if opt in pack.keys():
                        s += '%s=%s, ' % (opt, pack[opt])
                s = s.rstrip(' ').rstrip(',')
                fout.write('%s)\n' % s)
            elif 'grid' in data.keys():
                pass
            elif 'place' in data.keys():
                pass
            else:
                logging.warning('No placement for %s' % name)
        else:
            logging.warning('No style for %s' % name)

    def optionsString(self, style):
        opts=''
        excludeKeys = ('pack', 'grid', 'place')
        if style and len(style.keys())>0:
            for key in style.keys():
                if key in excludeKeys: continue
                opt = '%s=%s' % (key, style[key])
                opts += ', %s' % opt
        return opts
    
    def outputObjectCreation(self, parentName, key, data, fout, prefix):
        name, tags = self.decodeKey(key)
        clsName = tags[0]
        compName = ','.join(tags)
        style = self.findConfigurationFor(compName)
        opts = self.optionsString(style)
        #
        placeIt = True
        if compName == 'tk.Menu,bar':
            fout.write(prefix+'%s = %s(%s%s)\n' % (name, clsName, parentName, opts) )
            fout.write(prefix+'self.root.config(menu=%s)\n' % name)
            placeIt = False
        elif compName == 'tk.Menu,cascade':
            fout.write(prefix+'%s = %s(%s%s)\n' % (name, clsName, parentName, opts) )
            fout.write(prefix+'%s.add_cascade(label="%s", menu=%s)\n' % (parentName, name, name))
            pass
            placeIt = False
        elif compName == 'tk.Menu':
            if data == 'command':
                fout.write(prefix+'%s.add_command(label="%s")\n' % (parentName, name))
            elif data == 'checkbutton':
                fout.write(prefix+'%s.add_checkbutton(label="%s")\n' % (parentName, name))
            elif data == 'radiobutton':
                fout.write(prefix+'%s.add_radiobutton(label="%s")\n' % (parentName, name))
            placeIt = False
        elif clsName == 'ttk.Label':
            label = name
            if data: label = data
            fout.write(prefix+'%s = %s(%s, text="%s"%s)\n' %\
                       (name, clsName, parentName, label, opts) )
        elif clsName == 'ttk.Button':
            label = name
            if data: label = data
            fout.write(prefix+'%s = %s(%s, text="%s"%s)\n' %\
                       (name, clsName, parentName, label, opts) )
        else:
            fout.write(prefix+'%s = %s(%s%s)\n' % (name, clsName, parentName, opts) )
        if placeIt:
            if style:
                self.outputPlacement(name, style, fout, prefix)
            else:
                logger.warning('No style for %s, use pack(fill=tk.BOTH, expand=True)' % name)
                style1 = { 'pack': { 'fill': 'tk.BOTH', 'expand': 'True' } }
                self.outputPlacement(name, style1, fout, prefix)
        pass

    def outputSubComponent(self, parentName, key, data, fout, prefix=''):
        name, tags = self.decodeKey(key)
        clsName = tags[0]
        compName = ','.join(tags)
        style = self.findConfigurationFor(compName)
        #
        self.outputObjectCreation(parentName, key, data, fout, prefix)
        #
        keys = self.keys(data)
        for key1 in keys:
            name1, tags1 = self.decodeKey(key1)
            if len(tags1)==0:
                logger.warning('No class name give in tags for %s' % name)
                continue
            clsName1 = tags1[0]
            data1 = self.child(data, key1)
            fout.write('\n')
            self.outputSubComponent(name, key1, data1, fout, prefix)
        #
        logger.info(clsName)
        if clsName == 'ttk.PanedWindow':
            fout.write('\n')
            for key1 in keys:
                name1, tags1 = self.decodeKey(key1)
                fout.write(prefix+'%s.add(%s)\n'% (name, name1))
        pass
    
    def outputComponent(self, key, data, fout, prefix=''):
        name, tags = self.decodeKey(key)
        clsName = tags[0]
        fout.write(prefix+'class %s(%s):\n' % (name, clsName))
        fout.write(prefix+'    def __init__(self, parent):\n')
        fout.write(prefix+'        super().__init__(parent)\n')
        fout.write(prefix+'        self.root = parent\n')
        fout.write(prefix+'        self.pack(fill=tk.BOTH, expand=True)\n')
        fout.write(prefix+'        self.buildGui()\n')
        fout.write(prefix+'\n')
        fout.write(prefix+'    def buildGui(self):\n')
        keys = self.keys(data)
        for key1 in keys:
            name1, tags1 = self.decodeKey(key1)
            if len(tags1)==0:
                logger.warning('No class name give in tags for %s' % name)
                continue
            clsName1 = tags1[0]
            data1 = self.child(data, key1)
            prefix2 = prefix + '        '
            self.outputSubComponent('self', key1, data1, fout, prefix2)
            fout.write('\n')
        fout.write(prefix+'        pass\n')
        fout.write(prefix+'    pass\n')
    pass



#!/usr/bin/env python3
#------------------------------------------------------------------------
# dslot: sddgen.py
#------------------------------------------------------------------------
import yaml
import json
import argparse
import logging
import sddgen

logger = logging.getLogger(__name__)

def parseArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input-file', dest='inputFile',
                        type=str, default='', 
                        help='Input filename (.yaml)')
    parser.add_argument('-o', '--output-file', dest='outputFile',
                        type=str, default='', 
                        help='Output filename')
    parser.add_argument('-l', '--logLevel', dest='logLevel',
                        type=str, default='INFO', 
                        help='Log level (ERROR|WARNING|INFO|DEBUG)')
    return parser.parse_args()

def logLevel(slevel):
    level = 'INFO'
    if slevel == 'ERROR':
        level = logging.ERROR
    elif slevel == 'WARNING':
        level = logging.WARNING
    elif slevel == 'INFO':
        level = logging.INFO
    elif slevel == 'DEBUG':
        level = logging.DEBUG
    else:
        logger.warning('Given log level %s is not valid, using INFO' % slevel)
    return level

def run(args):
    gen = None
    dtype = 'NOT_SPECIFIED'
    with open(args.inputFile, 'r') as f:
        print('ssgen = ', dir(sddgen))
        data = sddgen.ModelFile(args.inputFile)
        data.load()
        dtype = data.header.modelType
        logger.info(f'dtype = {dtype}')
        if dtype == 'HTML+CSS':
            logger.info('HTML+CSS generator')
            gen = sddgen.HtmlGenerator(modelFile=data)
        elif dtype == 'Tkinter':
            gen = sddgen.GuiTkGenerator(modelFile=data)
    if gen:
        logger.info(f'Generate output dtype={dtype}')
        gen.filenameIn = args.inputFile
        gen.generate(args.outputFile)
        #print(json.dumps(data, indent=2))
    else:
        logger.error(f'No generator found for data type {dtype}')
        
if __name__ == '__main__':
    args = parseArgs()

    level = logLevel(args.logLevel)
    logging.basicConfig(level=level,
                        format='%(name)-20s %(levelname)-8s %(message)s')
                        
    
    logger.info('Running sddgen.py')

    run(args)
    
    



#------------------------------------------------------------------------
# sddgen.guitk: guiComponents.py
#------------------------------------------------------------------------
import time
import logging
import tkinter as tk
from tkinter import ttk

logger = logging.getLogger(__name__)

def addScrollBars(widget, parent, xscroll=False, yscroll=False, layout='grid'):
    if yscroll:
        yscrollbar = ttk.Scrollbar(parent, 
                                   orient=tk.VERTICAL,
                                   command=widget.yview)
        widget['yscrollcommand'] = yscrollbar.set
        if layout == 'grid':
            yscrollbar.grid(row=0, column=1, sticky=tk.NS)
        elif layout == 'pack':
            yscrollbar.pack(side=tk.RIGHT, fill=tk.Y, expand=True)
    if xscroll:
        xscrollbar = ttk.Scrollbar(parent, 
                                   orient=tk.HORIZONTAL, 
                                   command=widget.xview)
        widget['xscrollcommand'] = xscrollbar.set
        if layout == 'grid':
            xscrollbar.grid(row=1, column=0, sticky=tk.EW)
        elif layout == 'pack':
            xscrollbar.pack(side=tk.BOTTOM, fill=tk.X, expand=True)
    return (xscrollbar, yscrollbar)

#------------------------------------------------------------------------
# AnalysisPanel
#------------------------------------------------------------------------
class AnalysisPanel(ttk.LabelFrame):
    def __init__(self, parent, values, text='Analysis'):
        super().__init__(parent, text=text)
        cframe = ttk.Frame(self, height=600)
        cframe.pack(side=tk.TOP, fill=tk.X)

        cframe.grid_columnconfigure(0, weight=1)
        cframe.grid_columnconfigure(1, weight=5)
        cframe.grid_columnconfigure(2, weight=1)

        selectionLabel = ttk.Label(cframe, text='Type: ')
        selectionLabel.grid(row=0, column=0, sticky=tk.NSEW)
        self.selection = ttk.Combobox(cframe, values=values)
        self.selection.grid(row=0, column=1, sticky=tk.NSEW)
        self.runButton = ttk.Button(cframe, text='Run')
        self.runButton.grid(row=0, column=2, sticky=tk.NSEW)
        nameLabel = ttk.Label(cframe, text='Name: ')
        nameLabel.grid(row=1, column=0, sticky=tk.NSEW)
        self.nameEntry = ttk.Entry(cframe)
        self.nameEntry.grid(row=1, column=1, sticky=tk.NSEW)
        
        self.properties = ttk.LabelFrame(self, text='Properties')
        self.properties.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        columns = ('parameter', 'value')
        fields = [FieldEntry('p1', 0.0),
                  FieldEntry('p2', 'hello')]
        self.propertiesFrame = PropertyGridFrame(self.properties)
        self.propertiesFrame.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        #addScrollBars(self.propertiesFrame, self.properties, True, True)
        self.propertiesFrame.setFields(fields)

#------------------------------------------------------------------------
# ImageControlPanel
#------------------------------------------------------------------------
class ImageControlPanel(ttk.Frame):
    def __init__(self, panel, analysis):
        super().__init__(panel, style='P1.TFrame')
        self.analysis = analysis
        self.comboEntries = []
        self.parameters = {}
        self.build()

    def build(self):
        n = self.analysis.nInputImages
        print('ImageControlPanel n = %d' % n)
        self.comboEntries.clear()
        self.parameters.clear()
        #
        if n >= 1 and n <= 10:
            for i in range(n):
                text = ''
                if len(self.analysis.inputImages)>i:
                    text = self.analysis.inputImages[i].name
                x = ComboEntryPanel(self, text)
                x.pack(anchor=tk.NW, fill=tk.X)#, expand=True)
                self.comboEntries.append(x)
        for k, v in self.analysis.parameters.items():
            x1 = FieldEntryPanel(self, k, str(v))
            x1.pack(side=tk.TOP, fill=tk.X)#, expand=True)
            print('Add parameter field')
            self.parameters[k] = v
        #
        run = ttk.Button(self, text='Run', width=10)
        run.pack(side=tk.LEFT, expand=True)

    pass

class ScrollableFrame(ttk.Frame):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.canvas = tk.Canvas(self, bg='ivory', width=300, height=300)
        xbar, ybar = addScrollBars(self.canvas, self, True, True, layout='none')
        xbar.pack(side=tk.BOTTOM, fill=tk.X)
        ybar.pack(side=tk.RIGHT, fill=tk.Y)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        #
        self.frame = ttk.Frame(self.canvas, style='canvas.TFrame')
        self.canvas.create_window(0, 0, window=self.frame, anchor=tk.NW)
        self.frame.bind('<Configure>', self.onFrameConfigure)
        self.frame.bind('<Button-1>', self.onFrameClick)
        self.canvas.bind('<Button-1>', self.onFrameClick)
        #
        for i in range(20):
            title = f'Label{i}'
            #label = ttk.Label(self.frame, text=title)
            #label.pack(side=tk.TOP, fill=tk.X, expand=True)

    def onFrameConfigure(self, event=None):
        canvas1 = event.widget.master
        canvas1.configure(scrollregion=canvas1.bbox('all'))

    def onFrameClick(self, event=None):
        widget = event.widget
        cw = widget.winfo_width()
        ch = widget.winfo_height()
        logger.info(f'widget {widget} w/h = ({cw}, {ch})')

    def clear(self):
        for x in self.frame.winfo_children():
            x.destroy()
            
    def frame(self):
        return self.frame

    pass
        
class GalleryPanel(ScrollableFrame):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

    def addImageFrame(self, image, title=''):
        frame = ttk.Frame(self.frame)
        frame.pack(side=tk.TOP, fill=tk.X, expand=True)
        label = ttk.Label(frame, text=title)
        label.pack(anchor=tk.NW, fill=tk.X, expand=True)
        label = ttk.Label(frame, image=image)
        label.pack(anchor=tk.NW, fill=tk.X, expand=True)
    
    def clear(self):
        for w in self.frame.winfo_children():
            w.destroy()
    pass
    



#-----------------------------------------------------------------------
# sddgen: htmlcss.py
#-----------------------------------------------------------------------
import logging

from .model import Node, Model, Components, Styles
from .generator import Generator

logger = logging.getLogger(__name__)

class HtmlDocument(Node):
    def __init__(self, node):
        super().__init__(node.key, node.value)
        for key in self.keys():
            x = self.child(key)
            if x:
                self.value[key] = HtmlElement(x)
        pass

class HtmlElement(Node):
    html5Tags = ('body', 'h1', 'h2', 'h3', 'h4', 'h5',
                 'ul', 'li', 'ol',
                 'table', 'tr', 'th', 'td',
                 )
    
    def __init__(self, node):
        super().__init__(node.key, node.value)
        self.htmlTag = 'div'
        if self.typeName in HtmlElement.html5Tags:
            self.htmlTag = self.typeName
        for key in self.keys():
            x = self.child(key)
            if x:
                self.value[key] = HtmlElement(x)
        pass

class CssEntry(Node):
    def __init__(self, node):
        super().__init__(node.key, node.value)
        pass

class HtmlComponents(Components):
    def __init__(self, node):
        super().__init__(node)
        elements = {}
        for k, v in self.value.items():
            x = HtmlElement(v)
            x.isComponent = True
            elements[k] = x
        self.value = elements

    def find(self, name):
        x = None
        if name in self.value.keys():
            x = self.value[name]
        return x
    
    pass
    
class CssStyles(Styles):
    def __init__(self, node):
        super().__init__(node)
    pass
    
class HtmlGenerator(Generator):
    def __init__(self, model=None, output='sddout'):
        super().__init__(model=model, output=output)
        self.header = self.data.header
        self.document = HtmlDocument(self.data.document)
        self.components = HtmlComponents(self.data.components)
        self.styles = CssStyles(self.data.styles)

    def generate(self):
        out = self.outputName
        fn1 = f'{out}.html'
        fn2 = f'{out}.css'
        f1 = open(fn1, 'w')
        f2 = open(fn2, 'w')
        self.writeHtml(f1, fn2)
        self.writeCss(f2)
        f1.close()
        f2.close()

    def writeHtml(self, fout, cssfn, prefix=''):
        title='???'
        fout.write(f'<html>\n')
        fout.write(f'  <head>\n')
        fout.write(f'    <meta charset="utf-8" />\n')
        fout.write(f'    <title>{title}</title>\n')
        fout.write(f'    <link rel="stylesheet" href="{cssfn}" />\n')
        fout.write(f'  </head>\n')
        fout.write(f'  <body>\n')
        for key in self.document.keys():
            logger.debug(f'Write node {key}')
            node = self.document.child(key)
            self.writeElement(fout, node, prefix='    ')
        fout.write(f'  </body>\n')
        fout.write(f'</html>\n')
        pass

    def writeElement(self, fout, node, prefix=''):
        n = node.nChildren()
        htmlTag = node.htmlTag
        tags1 = node.tags
        component = None
        logger.debug(f'{prefix}Element: {htmlTag} n={n} -> {node.typeName}')
        if node.typeName != '':
            component = self.components.find(node.typeName)
            if component:
                logger.debug(f'{prefix}Use component {node.typeName}')
                return self.writeElement(fout, component, prefix)
        #
        if node.isComponent:
            tags1 = [node.name] + tags1
        classList = ' '.join(tags1)
        logger.debug(f'{prefix}classlist={classList}')
        if classList != '':
            fout.write(f'{prefix}<{htmlTag} class="{classList}">\n')
        else:
            fout.write(f'{prefix}<{htmlTag}>\n')
        #
        if n>0:
            keys = node.keys()
            for key in keys:
                node2 = node.child(key)
                if node2:
                    self.writeElement(fout, node2, prefix+'  ')
        #
        fout.write(f'{prefix}</{htmlTag}>\n')

    def writeCss(self, fout):
        for key in self.styles.keys():
            cssEntry = self.styles[key]
            self.writeCssEntry(fout, cssEntry, '')
        pass
    
    def writeCssEntry(self, fout, node, prefix=''):
        name = node.name
        etype = ''
        if not name in HtmlElement.html5Tags:
            etype = '.'
        fout.write(f'{prefix}{etype}{name}' + '{\n')
        for key in node.keys():
            value = node[key].value
            fout.write(f'{prefix}  {key}: {value};\n')
        fout.write(f'{prefix}' + '}\n')
        pass
    



#------------------------------------------------------------------------
# sddgen: model.py
#------------------------------------------------------------------------
import os
import re
import yaml
import logging

logger = logging.getLogger(__name__)

# Custom key
class Key:
    def __init__(self, key):
        self.key = key
        self.name, self.options, self.tags = self.decode()

    def stringToValue(self, value):
        x = value
        success = False
        try:
            x = int(value)
            success = True
        except ValueError:
            pass
        if not success:
            try:
                x = float(value)
                success = True
            except ValueError:
                pass
        if not success:
            if value in ('true', 'True', 'TRUE'):
                x = True
            elif value in ('false', 'False', 'FALSE'):
                x = False
        return x

    def decodeOptions(self, kvs):
        options = {}
        re1 = re.compile('^(.+)=(.+)$')
        for kv in kvs:
            mg = re1.match(kv)
            if mg:
                key, value = mg.group(1), mg.group(2)
                options[key] = self.stringToValue(value)
        return options

    def decode(self):
        key = self.key
        #logger.info(f'Decode key: {key}')
        re0 = re.compile('^(.+)$')
        re1 = re.compile('^(.+)\[(.*)\]$')
        re2 = re.compile('^(.+)\{(.*)\}$')
        re12 = re.compile('^(.+)\[(.*)\]\{(.*)\}$')

        name = ''
        tags = []
        options = {}
        # a[]{}
        mg = re12.match(key)
        if mg:
            #logger.info(f'match 12: {mg.groups()}')
            name = mg.group(1)
            options = self.decodeOptions(mg.group(2).split(','))
            tags = mg.group(3).split(',')
            return (name, options, tags)
        # a[]
        mg = re1.match(key)
        if mg:
            #logger.info(f'match 1: {mg.groups()}')
            name = mg.group(1)
            options = self.decodeOptions(mg.group(2).split(','))
            return (name, options, tags)
        # a{}
        mg = re2.match(key)
        if mg:
            #logger.info(f'match 2: {mg.groups()}')
            name = mg.group(1)
            tags = mg.group(2).split(',')
            return (name, options, tags)
        # a
        mg = re0.match(key)
        if mg:
            #logger.info(f'match 0: {mg.groups()}')
            name = mg.group(1)
            return (name, options, tags)
        return (name, options, tags)

# Node with tags and options
class Node:
    def __init__(self, data, key=None):
        self.data = data
        self.key = key
        self.name = ''
        self.tags = ''
        #
        if key and key!='':
            key1 = Key(key)
            self.name, self.options, self.tags = key1.decode()
        self.checkContent()

    def checkContent(self):
        if self.isScalarType():
            if self.isStringType():
                self.typeName = self.data
                
    def allTags(self):
        tags = list(self.tags)
        keys = self.keys()
        for k in keys:
            c = self.childNode(k)
            if c:
                for tag in c.tags:
                    if tag not in tags:
                        tags.append(tag)
        return tags

    def optionKeys(self):
        return self.options.keys()
    
    def allOptionKeys(self):
        okeys = self.options.keys()
        keys = self.keys()
        for k in keys:
            c = self.childNode(k)
            if c:
                okeys2 = c.optionKeys()
                for k2 in okeys2:
                    if k2 not in okeys:
                        okeys.append(k2)
        return okeys

    def allOptionValuesOf(self, optionKey):
        values = []
        if optionKey in self.options.keys():
            values.append(self.options[optionKey])
        keys = self.keys()
        for k in keys:
            c = self.childNode(k)
            if c:
                values2 = c.allOptionValuesOf(optionKey)
                for v2 in values2:
                    if v2 not in values:
                        values.append(v2)
        return values

    def nChildren(self):
        n = 0
        if self.isContainerType():
            n = len(self.data)
        return n
    
    def keys(self):
        keys = []
        if self.isObjectType():
            tag1 = '_orderedKeys'
            keys0 = self.data.keys()
            if tag1 in keys0:
                keys1 = self.data[tag1]
                name2key = {}
                for k in keys0:
                    key1 = Key(k)
                    name, options, tags = key1.decode()
                    if name in keys1:
                        name2key[name] = k
                for k in keys1:
                    keys.append(name2key[k])
            else:
                keys = list(filter(lambda x: not x.startswith('_'), self.data.keys()) )
            print('KEYS=', keys)
        elif self.isArrayType():
            keys = list(range(self.nChildren()))
        return keys
    
    def setValue(self, value):
        if type(value).__name__ == 'str' and \
           self.typeName == '': # and len(self.tags) == 0:
            self.typeName = value
        self.data = value

    def dataType(self):
        return type(self.data).__name__
    
    def isScalarType(self):
        return self.dataType() in ('int', 'float', 'str', 'bool')
    
    def isIntegerType(self):
        return self.dataType() == 'int'
    
    def isFloatType(self):
        return self.dataType() == 'float'
    
    def isStringType(self):
        return self.dataType() == 'str'
    
    def isBoolType(self):
        return self.dataType() == 'bool'
    
    def isArrayType(self):
        return self.dataType() == 'list'
    
    def isObjectType(self):
        return self.dataType() == 'dict'

    def isContainerType(self):
        return self.isArrayType() or self.isObjectType()

    def childNode(self, key):
        return Node(self[key], key)

    def items(self):
        if self.isObjectType():
            return self.data.items()
        else:
            return []
        
    def __getitem__(self, key):
        if self.isArrayType() or self.isObjectType():
            if self.isObjectType() and key not in self.data.keys():
                logger.error(f'Key error {key}, available={self.data.keys()}')
            return self.data[key]
        else:
            return None
    pass

def createNode(m, key=''):
    '''Create a Node from a Python object'''
    vnode = Node(key)
    tm = type(m)
    if tm == type(0) or tm == type(1.0) or tm == type(''):
        vnode.setValue(m)
    elif tm == type([]):
        value = []
        for i, node in enumerate(m):
            x = createNode(node, i)
            value.append(x)
        vnode.setValue(value)
    elif tm == type({}):
        value = {}
        for k, v in m.items():
            x = createNode(v, k)
            value[x.name] = x
            vnode.setValue(value)
    return vnode

class Header(Node):
    def __init__(self, data):
        super().__init__(data)
        #
        self.name = ''
        self.version = ''
        self.modelType = ''
        self.author = ''
        self.date = None
        self.setHeader(data)
        
    def setHeader(self, headerData):
        keys = headerData.keys()
        if 'name' in keys:
            self.name = headerData['name']
        if 'version' in keys:
            self.version = headerData['version']
        if 'modelType' in keys:
            self.modelType = headerData['modelType']
        if 'author' in keys:
            self.author = headerData['author']
        if 'date' in keys:
            dtstring = headerData['date']
            try:
                self.date = datetime.datetime.fromisoformat(dtstring)
            except ValueError:
                logger.error(f'{dtstring} does not look like a datetime format')
        
    def documentKey(self):
        key = 'document'
        if 'documentKey' in self.data.keys():
            key = self.data['documentKey']
        return key
    
    def componentsKey(self):
        key = 'components'
        if 'componentsKey' in self.data.keys():
            key = self.data['componentsKey']
        return key
    
    def stylesKey(self):
        key = 'styles'
        if 'stylesKey' in self.data.keys():
            key = self.data['stylesKey']
        return key

class ModelFile:
    def __init__(self, fileName=''):
        self.filePath = fileName
        self.header = None
        self.document = None
        self.documentKey = ''
        self.components = None
        self.styles = None

    def load(self):
        self.data = None
        if os.path.exists(self.filePath):
            with open(self.filePath, 'r') as f:
                self.data = yaml.load(f, Loader=yaml.SafeLoader)
        if self.data == None:
            logger.error(f'Cannot load sdd model from {self.filePath}')
            return
        #
        if 'header' in self.data.keys():
            self.header = Header(self.data['header'])
        if self.header:
            self.readContents()

    def readContents(self):
        topKeys = self.data.keys()
        dkey = self.header.documentKey()
        for key in topKeys:
            node = Node(None)
            key1 = Key(key)
            name, options, tags = key1.decode()
            if name == dkey:
                self.document = self.data[key]
                self.documentKey = key
                break
        key = self.header.componentsKey()
        if key in topKeys:
            self.components = self.data[key]
        key = self.header.stylesKey()
        if key in topKeys:
            self.styles = self.data[key]

    def findComponent(self, name):
        x = None
        for k, c in self.components.items():
            ckey = Key(k)
            if ckey.name == name:
                x = Node(c, k)
                break
        return x
            
    def findStyle(self, name):
        x = None
        for k, s in self.styles.items():
            skey = Key(k)
            if skey.name == name:
                x = Node(s, k)
                break
        return x

    def styleOfComponent(self, component):
        style = self.findStyle(component.name)
        return style
    
    def documentNode(self):
        logger.info(f'documentkey: {self.documentKey}')
        return Node(self.document, self.documentKey)
    
    def componentsNode(self):
        logger.info(f'CKEY:{self.header.componentsKey()}')
        return Node(self.components, self.header.componentsKey())
        
    def stylesNode(self):
        return Node(self.styles, self.header.stylesKey)
    
    def componentKeysInOrder(self):
        v = self.componentsNode()
        v2 = []
        keys = v.keys()
        for k in keys:
            c = v.childNode(k)
            if c==None:
                v2.insert(0, k)
                continue
            tags = c.allTags()
            imatch = -1
            for tag in tags:
                for i2, cname2 in enumerate(v2):
                    if cname2 == tag and i2 > imatch:
                        imatch = i2
            if imatch >= 0:
                v2.insert(imatch+1, k)
            else:
                v2.insert(0, k)
        return v2



#-----------------------------------------------------------------------
# sddgen: guitk.py
#-----------------------------------------------------------------------
import logging
from ..model import Key, Node, ModelFile
from .guiComponents import ScrollableFrame, addScrollBars

logger = logging.getLogger(__name__)

class GuiTkGenerator:
    def __init__(self, modelFile):
        self.modelFile = modelFile
        self.topComponentType = ''
        self.menuStack = []
        pass

    def writeWidget(self, fout, node, parentNode=None, prefix=''):
        parentName = None
        if parentNode: parentName = parentNode.name
        
        fout.write(f'{prefix}{node.name} = {node.tags}({parentName})\n')
        for key in node.keys():
            if key.startswith('_'): continue
            x = node.childNode(key)
            self.writeWidget(fout, x, node, prefix+'  ')
            
    def writeStyles(self, fout, style, prefix):
        pass

    def writePlacement(self, fout, component, style, prefix):
        if style == None: return
        keys = style.keys()
        if 'pack' in keys:
            pack = style.childNode('pack')
            if pack:
                pack = pack.data
                s = f'{component.name}.pack('
                if 'fill' in pack.keys(): s += f'fill={pack["fill"]},'
                if 'side' in pack.keys(): s += f'side={pack["side"]},'
                if 'anchor' in pack.keys(): s += f'anchor={pack["anchor"]},'
                if 'expand' in pack.keys(): s += f'expand={pack["expand"]},'
                if s[-1]==',': s = s[:-1]
                s += ')\n'
                fout.write(f'{prefix}{s}')
        elif 'grid' in keys:
            grid = style.childNode('grid')
            if grid:
                grid = grid.data
                s = f'{component.name}.grid('
                if 'row' in grid.keys(): s += f'row={grid["row"]},'
                if 'column' in grid.keys(): s += f'column={grid["column"]},'
                if 'weight' in grid.keys(): s += f'weight={grid["weight"]},'
                if 'sticky' in grid.keys(): s += f'sticky={grid["sticky"]},'
                if s[-1]==',': s = s[:-1]
                s += ')\n'
                fout.write(f'{prefix}{s}')
        pass

    def writeClassDefLine(self, fout, component, prefix):
        superClasses = ''
        if len(component.tags)>0:
            superClasses = f'({component.tags[0]}'
            for tag in component.tags[1:]:
                superClasses += f',{tag}'
            superClasses += ')'
        fout.write(f'{prefix}class {component.name}{superClasses}:\n')

    def valueCheckTk(self, value):
        x = value
        vtype = value.__class__.__name__
        if vtype == 'str' and \
           not (value.startswith('tk.') or value.startswith('ttk.') ):
            x = f'"{value}"'
        return x
    
    def buildInitArgs(self, component, style):
        args = ''
        clsname = component.tags[0]
        options = component.options
        if clsname == 'ttk.PanedWindow':
            logger.info(f'PanedWindow {component.options}')
            if 'orient' in component.options.keys():
                orient = component.options['orient']
                if not orient.startswith('tk.'): orient = f'"{orient}"'
                args += f', orient={orient}'
        argkeys1 = ('text', 'width', 'height')
        argkeys2 = ('text', 'width', 'height', 'style')
        for k, v in component.options.items():
            if k in argkeys1:
                args += f',{k}={self.valueCheckTk(v)}'
        if style and style.data:
            for k, v in style.data.items():
                if k in argkeys2:
                    args += f',{k}={self.valueCheckTk(v)}'
        return args
    
    def buildAddArgs(self, component, style):
        args = ''
        clsname = component.tags[0]
        options = component.options
        argkeys = ()
        if clsname == 'ttk.PanedWindow':
            argkeys = ('weight')
        #
        for k, v in component.options.items():
            if k in argkeys:
                args += f',{k}={self.valueCheckTk(v)}'
        if style and style.data:
            for k, v in style.data.items():
                if k in argkeys:
                    args += f',{k}={self.valueCheckTk(v)}'
        return args
    
    def writeInstanciate(self, fout, component, parentName, prefix):
        clsname = ''
        style = self.modelFile.styleOfComponent(component)
        if len(component.tags)>0:
            clsname = component.tags[0]
        args = self.buildInitArgs(component, style)
        logger.info(f'{component.name} to instanciate')
        fout.write(f'{prefix}{component.name} = {clsname}({parentName}{args})\n')
        fout.write(f'{prefix}self.{component.name} = {component.name}\n')

    def writeMenuBar(self, fout, menuBar, prefix):
        name = 'menuBar'
        clsname = menuBar.tags[0]
        logger.info(f'{name} N={len(self.menuStack)}')
        fout.write(f'{prefix}root = self.master.master\n')
        fout.write(f'{prefix}menuBar = self\n')
        fout.write(f'{prefix}root.config(menu=self)\n')
        fout.write(f'{prefix}self.menuBar = menuBar\n')

    def writeMenu(self, fout, menu, parentName, prefix):
        name = menu.name
        clsname = menu.tags[0]
        logger.info(f'{name} N={len(self.menuStack)}')
        if len(self.menuStack) == 0:
            fout.write(f'{prefix}self.{name} = {name}\n')
            pass
        elif len(self.menuStack) == 1:
            parentName = self.menuStack[-1]
            fout.write(f'{prefix}{name} = {clsname}({parentName}, tearoff=False)\n')
            fout.write(f'{prefix}{parentName}.add_cascade(label="{name}", menu={name})\n')
            fout.write(f'{prefix}self.{name} = {name}\n')
        else:
            parentName = self.menuStack[-1]
            fout.write(f'{prefix}{parentName}.add_command(label="{name}")\n')

    def checkScrollBars(self, fout, component, parentName, style, prefix):
        xscroll, yscroll = False, False
        if style and style.data:
            skeys = style.data.keys()
            if 'xyscroll' in skeys or 'xscroll' in skeys:
                xscroll = True
            if 'xyscroll' in skeys or 'yscroll' in skeys:
                yscroll = True
        skeys = component.options.keys()
        if 'xyscroll' in skeys or 'xscroll' in skeys:
            xscroll = True
        if 'xyscroll' in skeys or 'yscroll' in skeys:
            yscroll = True
        if xscroll or yscroll:
            fout.write(f'{prefix}{parentName}.rowconfigure(0, weight=1)\n')
            fout.write(f'{prefix}{parentName}.columnconfigure(0, weight=1)\n')
            fout.write(f'{prefix}sddgen.guitk.addScrollBars({component.name}, {parentName}, True, True)\n')
            
    def writeSubComponents(self, fout, component, parentName, prefix):
        ckeys = component.keys()
        # - Create subcomponents
        fout.write(f'{prefix}# create subcomponents of {component.name}\n')
        for k2 in ckeys:
            c2 = component.childNode(k2)
            if c2.tags[0] == 'tk.Menu':
                logger.info(f'Menu {component.name} n in stack {len(self.menuStack)}')
                self.writeMenu(fout, c2, parentName, prefix)
            else:
                self.writeInstanciate(fout, c2, parentName, prefix)
        # - Placement
        for k2 in ckeys:
            c2 = component.childNode(k2)
            style2 = self.modelFile.styleOfComponent(c2)
            self.checkScrollBars(fout, c2, parentName, style2, prefix)
            self.writePlacement(fout, c2, style2, prefix)
        # - add(...) for certain types of components
        if component.tags[0] == 'ttk.PanedWindow':
            logger.info(f'Add sub widget to PanedWindow')
            for k2 in ckeys:
                c2 = component.childNode(k2)
                style = self.modelFile.styleOfComponent(c2)
                args = self.buildAddArgs(c2, style)
                fout.write(f'{prefix}{component.name}.add({c2.name}{args})\n')

        # - Create sub-components recursively
        #fout.write('\n')
        for k2 in ckeys:
            c2 = component.childNode(k2)
            if c2.nChildren() > 0:
                fout.write('\n')
                if c2.tags[0] == 'tk.Menu':
                    logger.info(f'Add menu to stack N(before): {len(self.menuStack)}')
                    self.menuStack.append(c2.name)
                self.writeSubComponents(fout, c2, c2.name, prefix)
                if c2.tags[0] == 'tk.Menu':
                    logger.info(f'Add menu to stack N(after): {len(self.menuStack)}')
                    self.menuStack.pop()
        pass
    
    def writeClassDef(self, fout, component, prefix):
        style = self.modelFile.styleOfComponent(component)
        self.writeClassDefLine(fout, component, prefix)
        fout.write(f'{prefix}  def __init__(self, parent):\n')
        fout.write(f'{prefix}    super().__init__(parent)\n')
        if component.name == self.topNode.tags[0]:
            fout.write(f'{prefix}    self.root = parent\n')
        fout.write(f'{prefix}    self.buildGui()\n')
        if component.name == self.topNode.tags[0]:
            fout.write(f'{prefix}    self.pack(fill=tk.BOTH, expand=True)\n')
        fout.write(f'{prefix}    pass\n\n')
        
        # buildGui()
        fout.write(f'{prefix}  def buildGui(self):\n')
        if component.tags[0] == 'tk.Menu':
            self.writeMenuBar(fout, component, prefix+'    ')
            self.menuStack.append('menuBar')
            
        # - Create sub-components
        ckeys = component.keys()
        self.writeSubComponents(fout, component, 'self', prefix+'    ')
        #self.writeStyles(fout, style, prefix+'  ')
        fout.write(f'{prefix}    pass\n\n')

        if component.tags[0] == 'tk.Menu':
            self.menuStack.pop()

    def generateClassDefs(self, fout):
        logger.info('Generate class definitions')
        components = self.modelFile.componentsNode()
        keys = self.modelFile.componentKeysInOrder()
        for ckey in keys:
            key = Key(ckey)
            content = components.childNode(ckey)
            logger.info(f'{key.name} {key.tags}')
            self.writeClassDef(fout, content, '')

    def generate(self, outputFN):
        self.topNode = self.modelFile.documentNode()
        with open(outputFN, 'w') as fout:
            # Python imports
            fout.write('import tkinter as tk\n')
            fout.write('from tkinter import ttk\n')
            fout.write('import sddgen\n')
            fout.write('from .guiComponents import *\n')
            fout.write('\n')
            
            # Components
            logger.info('Generate class definitions')
            self.generateClassDefs(fout)
            
            # main section
            topComponent = self.modelFile.findComponent(self.topNode.tags[0])
            style = self.modelFile.styleOfComponent(topComponent)
            logger.info(f'Top node {self.topNode.name} tags: {self.topNode.tags}')
            fout.write('if __name__ == "__main__":\n')
            fout.write('  root = tk.Tk()\n')
            if style:
                if 'title' in style.data.keys():
                    fout.write(f'  root.title("{style.data["title"]}")\n')
                if 'geometry' in style.data.keys():
                    fout.write(f'  root.geometry("{style.data["geometry"]}")\n')
            fout.write(f'  style = ttk.Style()\n')
            fout.write(f'  style.configure("r.TFrame", background="red")\n')
            fout.write(f'  style.configure("g.TFrame", background="green")\n')
            fout.write(f'  style.configure("b.TFrame", background="blue")\n')
            fout.write(f'  {self.topNode.name} = {self.topNode.tags[0]}(root)\n')
            fout.write('  root.mainloop()\n')

